"""
Common ETL specific utilities and methods
"""
import string
import time
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from random import random
from typing import Any, List, TypeVar, Generic, Optional, Dict

from cassandra.cqlengine import columns, ValidationError
from cassandra.cqlengine.management import sync_table
from cassandra.cqlengine.models import Model
from cassandra.cqlengine.query import LWTException
from pandas import DataFrame
from hip_data_tools.common import LOG


class EtlStates(Enum):
    """Enumerator for the possible states of an ETL"""
    Ready = 'ready'
    Processing = 'processing'
    Succeeded = 'succeeded'
    Failed = 'failed'


class EtlSinkRecordState(Model):
    """Cassandra ORM model for the Etl Sink States"""
    etl_signature = columns.Text(primary_key=True, partition_key=True, required=True)
    record_identifier = columns.Text(primary_key=True, partition_key=True, required=True)
    record_state = columns.Text(required=True, default=EtlStates.Ready.value)
    state_created = columns.DateTime(required=True, default=datetime.now())
    state_last_updated = columns.DateTime(required=True, default=datetime.now())


def sync_etl_state_table():
    """
    Utility method to sync (Create) the table as per ORM model
    Returns: None
    """
    LOG.debug("Sinking Cassandra Table using model")
    sync_table(EtlSinkRecordState)


def _state_change_validation(current: EtlStates, new: EtlStates) -> None:
    if new in (EtlStates.Succeeded, EtlStates.Failed) and current != EtlStates.Processing:
        raise ValidationError(
            f"Current State '{current}' cannot transition to '{new}'")

    if new == EtlStates.Processing and current != EtlStates.Ready:
        raise ValidationError(
            f"Current State '{current}' cannot transition to '{new}'")

    if new == current:
        raise ValidationError(
            f"Current State '{current}' cannot transition to '{new}', "
            f"you might have duplicate records in your data set")

    if current == EtlStates.Succeeded:
        raise ValidationError(
            f"Current State '{current}' does not allow any transitions")


class EtlSinkRecordStateManager:
    """
    The Generic ETL Sink State manager that manages and persists the state of a record
    Args:
        record_identifier (str): A unique Identifier string to identify the sink record
        etl_signature (str): The Unique ETL Signature to identify the ETL
    """

    def __init__(self, record_identifier: str, etl_signature: str):
        self.record_identifier = record_identifier
        self.etl_signature = etl_signature
        self.remote_state = self._get_or_create_state()

    def _get_or_create_state(self):

        try:

            LOG.debug("Creating ETL record sink using etl_signature: "
                      "%s, record_identifier: %s",
                      self.etl_signature,
                      self.record_identifier)

            return EtlSinkRecordState.if_not_exists().create(
                etl_signature=self.etl_signature,
                record_identifier=self.record_identifier)

        except LWTException as e:

            LOG.debug("LWTException raised: \n %s", e)

            return EtlSinkRecordState.get(
                etl_signature=e.existing['etl_signature'],
                record_identifier=e.existing['record_identifier'])

    def _change_state(self, new_state):

        _state_change_validation(self.current_state(), new_state)

        self.remote_state.record_state = new_state.value
        self.remote_state.state_last_updated = datetime.now()
        self.remote_state.save()

    def current_state(self) -> EtlStates:
        """
        Get current state of the sid record
        Returns (EtlStates): current state

        """
        return EtlStates(self.remote_state.record_state)

    def succeeded(self) -> None:
        """
        Mark Record as succeeded
        Returns: None
        """
        self._change_state(EtlStates.Succeeded)

    def failed(self) -> None:
        """
        Mark Record as failed
        Returns: None
        """
        self._change_state(EtlStates.Failed)

    def processing(self) -> None:
        """
        Mark Record as processing
        Returns: None
        """
        self._change_state(EtlStates.Processing)

    def ready(self) -> None:
        """
        Mark Record as ready
        Returns: None
        """
        self._change_state(EtlStates.Ready)


def current_epoch() -> int:
    """
    Get the current epoch to millisecond precision
    Returns: int
    """
    return int(round(time.time() * 1000))


def get_random_string(length: int) -> str:
    """
    A random ascii lowercase string of certain length
    Args:
        length (int): length of the random string
    Returns: str
    """
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(length))


@dataclass
class SourceSettings:
    """ Abstract base dataclass for source settings """
    pass


class Extractor(ABC):
    """
    Abstract base class to define the functionalities of an Extractor

    Args:
        settings (SourceSettings): settings used to connect to a source
    """

    def __init__(self, settings: SourceSettings):
        self._settings = settings
        pass

    @abstractmethod
    def extract_next(self) -> Any:
        """
        Extracts a single datapoint

        Returns: Any

        """
        pass

    @abstractmethod
    def has_next(self) -> bool:
        """
        Checks if the extractor has any more data points

        Returns: bool

        """
        pass

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the state of the Extractor, usually reverting the state to its initial position

        Returns: None

        """
        pass


FromDataElementType = TypeVar('FromDataElementType')
ToDataElementType = TypeVar('ToDataElementType')


class Transformer(ABC, Generic[FromDataElementType, ToDataElementType]):
    """Abstract Base class for handling data transformations"""

    @abstractmethod
    def transform(self, data: FromDataElementType) -> ToDataElementType:
        """
        Transform the data element provided

        Args:
            data (Any): A data Element

        Returns: Any

        """
        pass


class DictTransformer(Transformer[Dict, Dict]):
    """ Abstract base class to handle dict to dict transformation """

    @abstractmethod
    def transform(self, data: Dict) -> Dict:
        """
        Transform the data element provided

        Args:
            data (Any): A data Element

        Returns: Any

        """
        pass


class DataFrameTransformer(Transformer[DataFrame, DataFrame]):
    """ Abstract base class to handle DataFrame to DataFrame transformation """

    @abstractmethod
    def transform(self, data: DataFrame) -> DataFrame:
        """
        Transform the data element provided

        Args:
            data (Any): A data Element

        Returns: Any

        """
        pass


@dataclass
class SinkSettings:
    """Dataclass to encapsulate settings to connect to and write to a data sink"""
    pass


class Loader(ABC):
    """
    Abstract Base class for defining Loaders that write data to sinks

    Args:
        settings (SinkSettings): setting to connect to the sink
    """
    def __init__(self, settings: SinkSettings):
        self._settings = settings
        pass

    @abstractmethod
    def load(self, data: Any) -> None:
        """
        Load a given data point onto the sink

        Args:
            data (Any): Any single data point

        Returns: None

        """
        pass


class ETL(ABC):
    """
    Base ETL class that defines the interaction of the three components of an ETL
    Args:
        extractor (Extractor): source data extractor
        transformers (List[Transformer]): series of transformation
        loader (Loader): loader to write data into a sink
    """

    def __init__(self, extractor: Extractor, loader: Loader,
                 transformers: Optional[List[Transformer]] = None):
        self.extractor = extractor
        if not transformers:
            transformers = []
        self.transformers = transformers
        self.loader = loader

    def has_next(self) -> bool:
        """
        Does the Source have any more data elements

        Returns (bool): True if data exists

        """
        return self.extractor.has_next()

    def execute_next(self) -> None:
        """
        Execute the Extraction, transformation and loading of the next data element from the source

        Returns: None

        """
        extracted_data = self.extractor.extract_next()
        transformed_data = extracted_data
        for transformer in self.transformers:
            transformed_data = transformer.transform(transformed_data)
        self.loader.load(transformed_data)

    def execute_all(self) -> None:
        """
        Execute the Extraction, transformation and loading of all data element from the source

        Returns: None

        """
        while self.has_next():
            self.execute_next()

    def reset_source(self) -> None:
        """
        Reset state of the source Loader, this will cause the loader to forget the items that have
        been loaded already and start afresh

        Returns:

        """
        self.extractor.reset()



"""
handle ETL of offline conversion data from Athena to Google Ads API
"""
from typing import List, Tuple

from attr import dataclass
from cassandra.cqlengine import ValidationError
from pandas import DataFrame

from hip_data_tools.apache.cassandra import (
    CassandraConnectionSettings,
    CassandraConnectionManager,
)
from hip_data_tools.common import LOG
from hip_data_tools.etl.athena_to_dataframe import (
    AthenaToDataFrame,
    AthenaToDataFrameSettings,
)
from hip_data_tools.etl.common import (
    EtlSinkRecordStateManager, sync_etl_state_table, EtlStates
)
from hip_data_tools.google.googleads import (
    GoogleAdsOfflineConversionUtil,
    GoogleAdsConnectionManager,
    GoogleAdsConnectionSettings,
    GoogleAdsClicksConversionUtil,
    GoogleAdsUploadClickConversionsRequestUtil,
)


@dataclass
class AthenaToGoogleAdsOfflineConversionSettings(AthenaToDataFrameSettings):
    """S3 to Cassandra ETL settings"""

    transformation_column_mapping: dict
    etl_identifier: str
    etl_state_manager_keyspace: str
    etl_state_manager_connection: CassandraConnectionSettings
    destination_batch_size: int
    destination_connection_settings: GoogleAdsConnectionSettings
    conversion_action_id: str
    customer_id: str


def _get_record_signature(record: dict):
    return f"{record['gclid']}||conversion_action||{record['conversion_date_time']}"


def _get_structured_issue(error, data):
    return {
        "error": error,
        "data": data,
    }


class AthenaToGoogleAdsOfflineConversion(AthenaToDataFrame):
    """
    Class to transfer parquet data from s3 to Cassandra
    Args:
        settings (AthenaToCassandraSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: AthenaToGoogleAdsOfflineConversionSettings):
        self.__settings = settings
        super().__init__(settings)
        self._googleads = None
        self._click_conversion = None
        self._upload_click_conversion_request = None

    def upload_next(self) -> (List[dict], List[dict], List[dict]):
        """
        Upload the next file in line from the athena table onto AdWords offline conversion
        Returns:
            verification_issues List[dict]: a tuple of lists outlining any
            verification failures
            successes List[dict]: The responses for successful uploads to
            the Google Adwords API
            failures List[dict]: The responses for failed uploads to the
            Google Adwords API
        """
        return self._process_data_frame(self.next())

    def upload_all(self) -> (List[dict], List[dict], List[dict]):
        """
        Upload all files from the Athena table onto AdWords offline conversion
        Returns:
            verification_issues List[dict]: a tuple of lists outlining any
            verification failures
            successes List[dict]: The responses for successful uploads to
            the Google Adwords API
            failures List[dict]: The responses for failed uploads to the
            Google Adwords API
        """

        verification_issues, successes, failures = [], [], []

        for key in self.list_source_files():
            issues, success, fail = self._process_data_frame(self.get_data_frame(key))

            verification_issues.extend(issues)
            successes.extend(success)
            failures.extend(fail)

        if verification_issues:
            LOG.warning("There were %s verification failures", len(verification_issues))

            LOG.debug("All verification failures: \n %s", verification_issues)

        if failures:
            LOG.warning(
                "There were %s failures uploading to the adwords " "API", len(failures)
            )

            LOG.info("Sample Failure: \n %s", failures[0])

            LOG.debug("All failures: \n %s", failures)

        LOG.info(
            "There were %s records successfully uploaded from a total of %s submitted items",
            len(successes),
            len(successes) + len(failures) + len(verification_issues),
        )

        return verification_issues, successes, failures

    def _get_googleads_util(self):
        if self._googleads is None:
            self._googleads = GoogleAdsOfflineConversionUtil(
                GoogleAdsConnectionManager(
                    self.__settings.destination_connection_settings
                )
            )
        return self._googleads

    def _get_googleads_click_conversion_util(self):
        if self._click_conversion is None:
            self._click_conversion = GoogleAdsClicksConversionUtil(
                GoogleAdsConnectionManager(
                    self.__settings.destination_connection_settings
                )
            )
        return self._click_conversion

    def _get_upload_click_conversion_request_util(self):
        if self._upload_click_conversion_request is None:
            self._upload_click_conversion_request = GoogleAdsUploadClickConversionsRequestUtil(
                GoogleAdsConnectionManager(
                    self.__settings.destination_connection_settings
                )
            )
        return self._upload_click_conversion_request

    def _get_sink_manager(self, record: dict) -> EtlSinkRecordStateManager:
        # Need to set up the cassandra connection
        return EtlSinkRecordStateManager(
            record_identifier=_get_record_signature(record),
            etl_signature=self.__settings.etl_identifier,
        )

    def _data_frame_to_destination_dict(self, data_frame: DataFrame) -> List[dict]:
        data_frame = data_frame.rename(
            columns=self.__settings.transformation_column_mapping
        )
        approved_fields = self._get_googleads_util().valid_fields
        drop_fields = [
            col for col in list(data_frame.columns) if col not in approved_fields
        ]
        data_frame = data_frame.drop(drop_fields, axis=1)
        return data_frame.to_dict("records")

    def _chunk_batches(self, lst: List[dict]) -> List[List[dict]]:
        n = self.__settings.destination_batch_size
        return [lst[i * n: (i + 1) * n] for i in range((len(lst) + n - 1) // n)]

    def _process_data_frame(self, data_frame) -> Tuple[list, list, list]:
        data_dict = self._data_frame_to_destination_dict(data_frame)
        self._state_manager_connect()
        ready_data, verification_issues = self._verify_data_before_upsert(data_dict)
        data_dict_batches = self._chunk_batches(ready_data)
        successes = []
        failures = []
        verification_issues = []

        for data_batch in data_dict_batches:
            data_to_process, processing_issue = self._mark_processing(data_batch)
            verification_issues.extend(processing_issue)
            click_conversions = [
                self._get_click_conversion_batch(dat) for dat in data_to_process
            ]
            request = [
                self._get_upload_conversion_request_batch(click_conversion)
                for click_conversion in click_conversions
            ]
            response = [self._upload_conversions(r) for r in request]
            success, fail = list(zip(*response))
            successes.extend(success)
            failures.extend(fail)

        return verification_issues, successes, failures

    def _upload_conversions(self, data_batch):
        return self._get_googleads_util().upload_conversions(data_batch)

    def _state_manager_connect(self):

        LOG.info("Connecting to Cassandra")

        conn = CassandraConnectionManager(self.__settings.etl_state_manager_connection)
        conn.setup_connection(self.__settings.etl_state_manager_keyspace)

        LOG.info("Cassandra connection established")

        sync_etl_state_table()

    def _mark_processing(self, data: List[dict]) -> (List[dict], List[dict]):
        data_for_processing = []
        issues = []
        for dat in data:
            try:
                self._get_sink_manager(dat).processing()
                data_for_processing.append(dat)
            except ValidationError as e:
                issues.append(_get_structured_issue(str(e), dat))
        return data_for_processing, issues

    def _mark_upload_results(self, fail: List[dict], success: List[dict]) -> None:
        for dat in success:
            self._get_sink_manager(dat).succeeded()
        for dat in fail:
            self._get_sink_manager(dat["data"]).failed()

    def _verify_data_before_upsert(self, data: List[dict]) -> (List[dict], List[dict]):
        data, issues = map(list, zip(*[self._sanitise_data(dat) for dat in data]))

        # Remove None from the List
        data_clean = [i for i in data if i]
        issues_clean = [i for i in issues if i]

        if len(issues) > 0:
            LOG.warning(f"Issues found in verification, number of issues: {len(issues_clean)}", )

        return data_clean, issues_clean

    def _sanitise_data(self, dat):
        try:

            current_state = self._get_sink_manager(dat).current_state()

            LOG.debug("Current state of sink manager %s", current_state)

            if current_state == EtlStates.Ready:
                LOG.debug("Record in ready state with data: %s", dat)

                return dat, None
            else:

                LOG.debug("Sink state found to be not ready, state is %s, the "
                          "data is: "
                          "%s", current_state, dat)

                return None, _get_structured_issue(f"Current state is {current_state} "
                                                   "state", dat)
        except ValidationError as e:
            LOG.warning("Issue while trying to ready a record for the upload \n %s \n %s", e,
                        dat)
            return None, _get_structured_issue(str(e), dat)

    def _get_click_conversion_batch(self, data):
        return self._get_googleads_click_conversion_util().click_conversion(data)

    def _get_upload_conversion_request_batch(self, click_conversion):
        return self._get_upload_click_conversion_request_util().upload_click_conversion(self.__settings.customer_id,
                                                                                        click_conversion
                                                                                        )



"""
S3 related ETL components Extractors, Loaders, Transformers, etc.

"""

from abc import abstractmethod
from dataclasses import dataclass
from typing import Optional, List, Tuple, Any, NewType

from pandas import DataFrame

from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.common import LOG
from hip_data_tools.etl.common import SourceSettings, Extractor, SinkSettings, Loader, Transformer

S3Key = NewType("S3Key", str)
FileName = NewType("FileName", str)
S3Bucket = NewType("S3Bucket", str)


@dataclass
class S3SourceSettings(SourceSettings):
    """ Data Class that holds source settings to connect to a s3 bucket """
    bucket: S3Bucket
    key_prefix: S3Key
    suffix: Optional[str]
    connection_settings: AwsConnectionSettings


class S3Extractor(Extractor):
    """
    Base class for all types of S3 Extractors which handles connection to s3

    Args:
        settings (S3SourceSettings):  settings to connect to the source s3 bucket
    """

    def __init__(self, settings: S3SourceSettings):
        super().__init__(settings)
        self._settings = settings
        self._s3_util = None

    def _get_s3_util(self) -> S3Util:
        if self._s3_util is None:
            self._s3_util = S3Util(
                bucket=self._settings.bucket,
                conn=AwsConnectionManager(self._settings.connection_settings),
            )
        return self._s3_util

    @abstractmethod
    def extract_next(self):
        pass

    @abstractmethod
    def has_next(self) -> bool:
        pass


class S3FilesExtractor(S3Extractor):
    """
    Base class that lists all source files and can be used downstream to extract various
    types of data from these files

    Args:
        settings (S3SourceSettings):  settings to connect to the source s3 bucket
    """

    def __init__(self, settings: S3SourceSettings):
        super().__init__(settings)
        self._source_keys = None
        self._file_counter = 0

    def list_source_files(self) -> List[S3Key]:
        """
        Lists all files that match the source settings

        Returns: List[S3Key]

        """
        if self._source_keys is None:
            keys = self._get_s3_util().get_keys(self._settings.key_prefix)
            if self._settings.suffix:
                keys = [key for key in keys if key.endswith(self._settings.suffix)]
            self._source_keys = keys
            LOG.info("Listed and cached %s source files", len(self._source_keys))
            self._file_counter = len(self._source_keys)
        return self._source_keys

    def _next_file_path(self) -> str:
        file = self.list_source_files()[self._file_counter - 1]
        self._file_counter -= 1
        return file

    @abstractmethod
    def extract_next(self) -> Tuple[str, str]:
        pass

    def has_next(self) -> bool:
        self.list_source_files()
        return self._file_counter > 0

    def reset(self) -> None:
        self._source_keys = None


class S3FileLocationExtractor(S3FilesExtractor):
    """
    Base class for S3 file name extractors

    Args:
        settings (S3SourceSettings): settings to connect to the source s3 bucket
    """

    def __init__(self, settings: S3SourceSettings):
        super().__init__(settings)

    def extract_next(self) -> Tuple[str, str]:
        """
        next eligible source coordinates
        Returns: Tuple[str,str]
        """
        return self._settings.bucket, self._next_file_path()


class S3ParquetFileDataExtractor(S3FilesExtractor):
    """
    Extractor class to read parquet files stored in s3 in form of Pandas Dataframe

    Args:
        settings (S3SourceSettings): settings to connect to the source s3 bucket
    """

    def __init__(self, settings: S3SourceSettings):
        super().__init__(settings)

    def extract_next(self) -> DataFrame:
        """
        next eligible source coordinates
        Returns: Tuple[str,str]
        """
        s3 = self._get_s3_util()
        return s3.download_parquet_as_dataframe(key=self._next_file_path())


@dataclass
class S3SinkSettings(SinkSettings):
    """ Target sink settings for S3 """
    bucket: S3Bucket
    connection_settings: AwsConnectionSettings


class S3Loader(Loader):
    """
    Base class for all S3 Loaders
    """

    def __init__(self, settings: S3SinkSettings):
        super().__init__(settings)
        self._settings = settings
        self._s3_util = None

    def _get_s3_util(self) -> S3Util:
        if self._s3_util is None:
            self._s3_util = S3Util(
                bucket=self._settings.bucket,
                conn=AwsConnectionManager(self._settings.connection_settings),
            )
        return self._s3_util

    @abstractmethod
    def load(self, data: Any) -> None:
        pass


class S3FileCopyLoader(S3Loader):
    """
    Class to copy files from various buckets to a target bucket

    Args:
        settings (S3SinkSettings): target s3 data sink settings
    """

    def __init__(self, settings: S3SinkSettings):
        super().__init__(settings)

    def load(self, data: Tuple[S3Bucket, S3Key, S3Key]) -> None:
        """
        Transfer one source s3 key to target
        Args:
            data (Tuple[S3Bucket, S3Key, S3Key]): source s3 bucket, source key and target key tuple
        Returns: None
        """
        (source_bucket, source_key, target_key) = data
        if not target_key:
            target_key = source_key
        s3 = self._get_s3_util().get_client()
        copy_source = {
            'Bucket': source_bucket,
            'Key': source_key
        }
        LOG.info("Transferring Key s3://%s/%s to s3://%s/%s",
                 source_bucket,
                 source_key,
                 self._settings.bucket,
                 target_key)
        s3.copy(copy_source, self._settings.bucket, target_key)


class S3DataFrameAsParquetFileLoader(S3Loader):
    """
    Class to Load Parquet files stored as dataframe into pandas dataframes
    Args:
        settings (S3SinkSettings): target s3 data sink settings
    """

    def __init__(self, settings: S3SinkSettings):
        super().__init__(settings)

    def load(self, data: Tuple[DataFrame, S3Key, FileName]) -> None:
        """
        Load the data that contains a dataframe, target s3 key and filename onto the sink as
        parquet files

        Args:
            data (Tuple[DataFrame, S3Key, FileName]): dataframe, target s3 key, target file name

        Returns: None

        """
        data_frame, key, file_name = data
        s3 = self._get_s3_util()
        s3.upload_dataframe_as_parquet(dataframe=data_frame, key=key, file_name=file_name)


class S3FileNameTransformer(Transformer[Tuple[S3Bucket, S3Key], Tuple[S3Bucket, S3Key]]):
    """ Abstract base class to handle Tuple[S3Bucket, S3Key] transformation """

    def __init__(self):
        pass

    @abstractmethod
    def transform(self, data: Tuple[S3Bucket, S3Key]) -> Tuple[S3Bucket, S3Key]:
        """
        Transform the data element provided

        Args:
            data (Any): A data Element

        Returns: Any

        """
        pass


class AddTargetS3KeyTransformer(S3FileNameTransformer):
    """
    Transformer to change the source prefix / directory of an s3 key into a given target
    directory, keeping file names same

    Args:
        target_key_prefix (S3Key): the new static target directory
    """

    def __init__(self, target_key_prefix: S3Key):
        super().__init__()
        self.target_key_prefix = target_key_prefix

    def transform(self, data: Tuple[S3Bucket, S3Key]) -> Tuple[S3Bucket, S3Key, S3Key]:
        """
        change the source prefix / directory of an s3 key into a given target directory,
        keeping file names same

        Args:
            data (Tuple[S3Bucket, S3Key]): Pair of source s3 bucket and key

        Returns: Tuple[S3Bucket, S3Key]

        """
        source_bucket, source_key = data
        file_name = source_key.split('/')[-1]
        target = None
        if self.target_key_prefix:
            target = S3Key(f"{self.target_key_prefix}/{file_name}")
        return source_bucket, source_key, target



"""
Module to deal with data transfer from Adwords to Athena
"""
from typing import List, Any, Optional, Tuple

from attr import dataclass
from pandas import DataFrame

from hip_data_tools.common import LOG
from hip_data_tools.aws.athena import AthenaUtil, get_athena_columns_from_dataframe, \
    extract_athena_type_from_value
from hip_data_tools.aws.common import AwsConnectionManager
from hip_data_tools.etl.adwords_to_s3 import AdWordsToS3Settings, AdWordsToS3, \
    AdWordsReportToS3Settings, AdWordsReportsToS3


@dataclass
class AdWordsToAthenaSettings(AdWordsToS3Settings):
    """Settings container for Adwords to Athena ETL"""
    target_database: str
    target_table: str
    target_table_ddl_progress: bool
    is_partitioned_table: bool
    partition_values: Optional[List[Tuple[str, Any]]]
    athena_columns: Optional[List[dict]]


class AdWordsToAthena(AdWordsToS3):
    """
    ETL Class to handle the transfer of data from adwords based on AWQL to S3 as parquet files
    Args:
        settings (AdWordsToS3Settings): the etl settings to be used
    """

    def __init__(self, settings: AdWordsToAthenaSettings):
        self.__settings = settings
        self.base_dir = settings.target_key_prefix
        if self.__settings.is_partitioned_table:
            partition_dirs = "/".join([f"{k}={v}" for k, v in settings.partition_values])
            settings.target_key_prefix = f"{settings.target_key_prefix}/{partition_dirs}"
        super().__init__(settings)

    def _get_athena_util(self):
        return AthenaUtil(
            database=self.__settings.target_database,
            conn=AwsConnectionManager(
                settings=self.__settings.target_connection_settings),
            output_bucket=self.__settings.target_bucket)

    def create_athena_table(self) -> None:
        """
        Creates an athena table on top of the transferred data
        Returns: None
        """
        self.build_query(start_index=0, page_size=1, num_iterations=1)
        data = self._get_next_page()
        athena_util = self._get_athena_util()
        if self.__settings.target_table_ddl_progress:
            athena_util.drop_table(self.__settings.target_table)
        athena_table_settings = self._construct_athena_table_settings(data)
        athena_util.create_table(table_settings=athena_table_settings)
        # Reset query state
        self.query = None

    def _construct_athena_table_settings(self, data: DataFrame) -> dict:
        partition_settings = []
        if self.__settings.is_partitioned_table:
            partition_settings = [{"column": k, "type": extract_athena_type_from_value(v)}
                                  for k, v in self.__settings.partition_values]
        athena_table_settings = {
            "exists": True,
            "partitions": partition_settings,
            "storage_format_selector": "parquet",
            "encryption": False,
            "table": self.__settings.target_table,
            "columns": self.__get_athena_columns(data),
            "s3_bucket": self.__settings.target_bucket,
            "s3_dir": self.base_dir,
        }
        return athena_table_settings

    def __get_athena_columns(self, data):
        if self.__settings.athena_columns:
            return self.__settings.athena_columns
        return get_athena_columns_from_dataframe(data)


@dataclass
class AdWordsReportsToAthenaSettings(AdWordsReportToS3Settings):
    """Settings container for Adwords to Athena ETL"""
    target_database: str
    target_table: str
    target_table_ddl_progress: bool
    is_partitioned_table: bool
    partition_values: Optional[List[Tuple[str, Any]]]


class AdWordsReportsToAthena(AdWordsReportsToS3):
    """
    ETL Class to handle the transfer of data from adwords based on AWQL to S3 as parquet files
    Args:
        settings (AdWordsToS3Settings): the etl settings to be used
    """

    def __init__(self, settings: AdWordsReportsToAthenaSettings):
        self.__settings = settings
        self.base_dir = settings.target_key_prefix
        if self.__settings.is_partitioned_table:
            partition_dirs = "/".join([f"{k}={v}" for k, v in settings.partition_values])
            settings.target_key_prefix = f"{settings.target_key_prefix}/{partition_dirs}"
        self._final_target_prefix = settings.target_key_prefix
        super().__init__(settings)

    def _get_athena_util(self):
        return AthenaUtil(
            database=self.__settings.target_database,
            conn=AwsConnectionManager(
                settings=self.__settings.target_connection_settings),
            output_bucket=self.__settings.target_bucket)

    def add_partitions(self):
        """
        Add the current Data Transfer's partition to Athena's Metadata
        Returns: None
        """
        if self.__settings.is_partitioned_table:
            athena_util = self._get_athena_util()
            athena_util.add_partitions(
                table=self.__settings.target_table,
                partition_keys=[key for (key, value) in self.__settings.partition_values],
                partition_values=[value for (key, value) in self.__settings.partition_values]
            )
        else:
            LOG.warning("The table is not partitioned, this is a NOOP")

    def create_athena_table(self) -> None:
        """
        Creates an athena table on top of the transferred data
        Returns: None
        """
        athena_util = self._get_athena_util()
        s3_util = self._get_s3_util()
        keys = s3_util.get_keys(key_prefix=self._final_target_prefix)
        LOG.debug("gathered files transferred under this ETL %s", keys)
        if keys:
            data = s3_util.download_parquet_as_dataframe(keys[0])
            LOG.info(
                "Downloaded parquet file from s3 to construct Athena create table statement: %s "
                "\n made dataframe of shape %s", keys[0], data.shape)
            if self.__settings.target_table_ddl_progress:
                athena_util.drop_table(self.__settings.target_table)
            athena_table_settings = self._construct_athena_table_settings(data)
            athena_util.create_table(table_settings=athena_table_settings)
        else:
            raise ValueError(
                "No Data has been uploaded to target directory, please load data first, "
                "before creating Athena table")

    def _construct_athena_table_settings(self, data: DataFrame) -> dict:
        partition_settings = []
        if self.__settings.is_partitioned_table:
            partition_settings = [{"column": k, "type": extract_athena_type_from_value(v)}
                                  for k, v in self.__settings.partition_values]
        athena_table_settings = {
            "exists": True,
            "partitions": partition_settings,
            "storage_format_selector": "parquet",
            "encryption": False,
            "table": self.__settings.target_table,
            "columns": get_athena_columns_from_dataframe(data),
            "s3_bucket": self.__settings.target_bucket,
            "s3_dir": self.base_dir,
        }
        return athena_table_settings

    def get_target_prefix_with_partition_dirs(self) -> str:
        """
        Return the target s3 key prefix which includes partition directories
        Returns: modified target key prefix string
        """
        return self.__settings.target_key_prefix



"""
Module contains variables and methods used for common / shared operations throughput the google
services package
"""
from abc import ABC, abstractmethod
from typing import List

from attr import dataclass
from oauth2client.service_account import ServiceAccountCredentials

from hip_data_tools.common import SecretsManager, ENVIRONMENT, KeyValueSource


class GoogleApiSecretsManager(SecretsManager):
    """

    Args:
        source (KeyValueSource): a kv source that has secrets
        key_json_var (str): The variable name or the key for finding the key file json object
    """

    def __init__(self,
                 source: KeyValueSource = ENVIRONMENT,
                 key_json_var: str = "key_json"):
        self._required_keys = [key_json_var, ]
        super().__init__(self._required_keys, source)

        self.key_json = self.get_secret(key_json_var)


@dataclass
class GoogleApiConnectionSettings:
    """Encapsulates the Google API connection settings"""
    secrets_manager: GoogleApiSecretsManager


class GoogleApiConnectionManager(ABC):
    """
    Google API connection manager abstract class
    """

    def __init__(self, settings: GoogleApiConnectionSettings, scope: List[str]):
        self.settings = settings
        self.scope = scope

    def _credentials(self):
        """
        Get the credentials for google sheets
        Returns (ServiceAccountCredentials): credentials object to authorize google sheet service
        """
        return ServiceAccountCredentials.from_json_keyfile_dict(
            self.settings.secrets_manager.key_json,
            self.scope)

    @abstractmethod
    def get_client(self):
        """
        Get the connection for google sheets
        Returns: authenticated client object of the connection library
        """



"""
This Module handles the connection and operations on Google AdWords accounts using adwords API
"""
import math
from collections import OrderedDict
from typing import List, Optional, Any
from attr import dataclass
from google.ads.googleads.client import GoogleAdsClient
from google.ads.googleads import oauth2
from pandas import DataFrame

from hip_data_tools.common import (
    KeyValueSource,
    ENVIRONMENT,
    SecretsManager,
    nested_list_of_dict_to_dataframe,
    LOG,
)


class GoogleAdsSecretsManager(SecretsManager):
    """
    Secrets manager for Google Adwords Specific secrets, these are modeled to handle the OAuth
    based google client and refresh token secrets
    Args:
        source (KeyValueSource): The Key Value source which holds secrets
        client_secret_var (str): The Key pointer for client_secret
        refresh_token_var (str): The Key pointer for refresh_token
        developer_token_var (str): The Key pointer for developer_token
    """

    def __init__(
        self,
        source: KeyValueSource = ENVIRONMENT,
        client_secret_var: str = "adwords_client_secret",
        refresh_token_var: str = "adwords_refresh_token",
        client_id_var: str = "adwords_client_id",
        client_customer_id_var: str = "adwords_client_customer_id",
        developer_token_var: str = "adwords_developer_token",
    ):
        self._required_keys = [
            client_secret_var,
            refresh_token_var,
            developer_token_var,
            client_id_var,
            client_customer_id_var,
        ]
        super().__init__(self._required_keys, source)
        self.client_secret = self.get_secret(client_secret_var)
        self.refresh_token = self.get_secret(refresh_token_var)
        self.developer_key = self.get_secret(developer_token_var)
        self.client_id = self.get_secret(client_id_var)
        self.client_customer_id = self.get_secret(client_customer_id_var)


@dataclass
class GoogleOAuthConnectionSettings:
    """
    Handle connection settings for AdWords client
    """

    client_id: str
    secrets_manager: GoogleAdsSecretsManager


class GoogleOAuthConnectionManager:
    """
    Creates an OAuth client from google ads library
    Args:
        settings (GoogleOAuthConnectionSettings): connection settings to use for creation
    """

    def __init__(self, settings: GoogleOAuthConnectionSettings):
        self.__settings = settings

    def _get_oauth_client(self) -> GoogleAdsClient:
        """
        Create an OAuth refresh-token client for all google ads and manager api
        Returns: GoogleAdsClient
        """
        return oauth2.get_installed_app_credentials(
            client_id=self.__settings.client_id,
            client_secret=self.__settings.secrets_manager.client_secret,
            refresh_token=self.__settings.secrets_manager.refresh_token,
        )


@dataclass
class GoogleAdsConnectionSettings(GoogleOAuthConnectionSettings):
    """
    Handle connection settings for AdWords client
    """

    user_agent: str
    client_customer_id: Optional[str]


class GoogleAdsConnectionManager(GoogleOAuthConnectionManager):
    """
    Google Ads Connection Manager that manages the lifecycle of authentication and the resulting
    GoogleAds client
    Args:`
        settings (GoogleAdWordsConnectionSettings): setting to use for connecting to adwords
    """

    def __init__(self, settings: GoogleAdsConnectionSettings):
        super().__init__(settings)
        self.settings = settings
        self._googleads_client = None

    def get_googleads_client(self, **kwargs) -> GoogleAdsClient:
        """
        Gets the connected google ads client, and creates one if not exist
        Returns (GoogleAdsClient): Google Ads Client

        """
        if self._googleads_client is None:
            self._googleads_client = self._create_client(**kwargs)
        return self._googleads_client

    def _create_client(self, **kwargs) -> GoogleAdsClient:
        oauth2_client = self._get_oauth_client()
        return GoogleAdsClient(
            developer_token=self.settings.secrets_manager.developer_key,
            credentials=oauth2_client,
            login_customer_id=self.settings.client_customer_id,
            **kwargs,
        )


class GoogleAdsUtil:
    """
    Generic Google Ads Utility class generates the required services to consume Google Ads API
    Args:
        conn (GoogleAdsConnectionManager): Connection manager to handle the creation of
        google ads client
    """

    def __init__(
        self,
        conn: GoogleAdsConnectionManager,
        service: str = None,
        version: str = None,
        type: str = None,
    ):
        self.service = service
        self.version = version
        self.type = type
        self.conn = conn
        self._service_object = None
        self._type_object = None
        self._pager = None
        self.query = None

    def _get_service(self, **kwargs):
        if not self._service_object:
            self._service_object = self._create_service(**kwargs)
        return self._service_object

    def _create_service(self, **kwargs):
        client = self.conn.get_googleads_client(**kwargs)
        return client.get_service(self.service, version=self.version)

    def _get_type(self, **kwargs):
        if not self._type_object:
            self._type_object = self._create_type(**kwargs)
        return self._type_object

    def _create_type(self, **kwargs):
        client = self.conn.get_googleads_client(**kwargs)
        return client.get_type(self.type, version=self.version)


class AdWordsDataReader(GoogleAdsUtil):
    """
    Generic Adwords Utility class generates the required services to consume adwords API
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(
        self,
        conn: GoogleAdsConnectionManager,
        service: str,
        version: str,
        page_size: int = 1000,
    ):
        super().__init__(conn, service, version)
        self.__pager = None
        self.query = None
        self.page_size = page_size

    def download_all_as_dict(self) -> List[dict]:
        """
        Generates list of Dict from the adwords query language query as described in adwords api
        Returns: List[dict]
        """
        complete_result = []
        for page in self._get_query_pager(self._get_query()):
            complete_result.extend(get_page_as_list_of_dict(page))
        return complete_result

    def download_all_as_dataframe(self) -> DataFrame:
        """
        Generates a data frame from the next page of the API call to adwords
        Returns: List[dict]
        """
        data_list = self.download_all_as_dict()
        return nested_list_of_dict_to_dataframe(data_list)

    def download_next_page_as_dict(self) -> List[dict]:
        """
        Generates a list of dict from the next page of the API call to adwords
        Returns: List[dict]
        """
        return get_page_as_list_of_dict(next(self._get_query_pager(self._get_query())))

    def download_next_page_as_dataframe(self) -> DataFrame:
        """
        Generates a data frame from the next page of the API call to adwords
        Returns: List[dict]
        """
        data_list = self.download_next_page_as_dict()
        return nested_list_of_dict_to_dataframe(data_list)

    def set_query(self, query) -> None:
        """
        Sets the new query for the class to use awql, once the query is set, the download methods
        can iterate over paged results
        Args:
            query (ServiceQueryBuilder): query built as per adwords query language
        Returns: None
        """
        self.__pager = None
        self.query = query

    def _get_query(self):
        if self.query is None:
            raise Exception(
                "Please set the query attribute using the method set_query(...)"
            )
        return self.query

    def _get_query_pager(self, query):
        if self.__pager is None:
            self.__pager = query.Pager(self._get_service())
        return self.__pager

    def get_parallel_payloads(
        self, page_size: int, number_of_workers: int
    ) -> List[dict]:
        """
        gives a list of dicts that contain start index, page size, and number of iterations
        Args:
            page_size (int): number of elements in each page / api call
            number_of_workers (int): total number of parallel workers for which the payload needs
            to be distributed
        Returns: List[dict] eg:
        [
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 0, 'worker': 0},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 393000, 'worker': 1},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 786000, 'worker': 2},
        ]
        """
        estimator = AdWordsParallelDataReadEstimator(
            conn=self.conn,
            service=self.service,
            version=self.version,
            query=self._get_query(),
        )
        return estimator.get_parallel_payloads(page_size, number_of_workers)


def _refine_result_list(
    number_of_workers: int, page_size: int, result: list, total_entries: int
) -> None:
    last_page_size = total_entries % page_size
    number_of_workers_with_non_zero_pages = math.ceil(total_entries / page_size)
    for workers_with_non_zero_pages in range(number_of_workers_with_non_zero_pages):
        result[workers_with_non_zero_pages]["number_of_pages"] = 1
        result[workers_with_non_zero_pages]["page_size"] = page_size
    for worker_with_zero_pages in range(
        number_of_workers_with_non_zero_pages, number_of_workers
    ):
        result[worker_with_zero_pages]["start_index"] = 0
    result[number_of_workers_with_non_zero_pages - 1]["page_size"] = last_page_size


def _refine_page_confings(
    page_size: int, pages_per_worker: int, total_entries: int, number_of_workers: int
) -> (int, int):
    page_size_per_worker = page_size
    refined_number_of_pages_per_worker = pages_per_worker
    if page_size >= total_entries:
        page_size_per_worker = 0
    elif page_size * number_of_workers > total_entries:
        page_size_per_worker = 0
        refined_number_of_pages_per_worker = 0
    return page_size_per_worker, refined_number_of_pages_per_worker


class AdWordsParallelDataReadEstimator(GoogleAdsUtil):
    """
    Class to use a service endpoint and awql query to estimate the number of parallel payloads, and
    their offsets
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
        service (str): Adwords service api to use for querying
        version (str): Adwords service api version to use for querying
        query (ServiceQueryBuilder): the Query builder object without limit clause
    """

    def __init__(
        self, conn: GoogleAdsConnectionManager, service: str, version: str, query
    ):
        super().__init__(conn, service, version)
        self.query = query

    def get_parallel_payloads(
        self, page_size: int, number_of_workers: int
    ) -> List[dict]:
        """
        gives a list of dicts that contain start index, page size, and number of iterations
        Args:
            page_size (int): number of elements in each page / api call
            number_of_workers (int): total number of parallel workers for which the payload needs
            to be distributed
        Returns: List[dict] eg:
        [
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 0, 'worker': 0},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 393000, 'worker': 1},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 786000, 'worker': 2},
        ]
        """
        total_entries = self._get_total_entries()
        last_page_residue = total_entries % page_size
        number_of_pages = int(total_entries / page_size)
        if last_page_residue > 0:
            number_of_pages += 1
        pages_per_worker = round(number_of_pages / number_of_workers)
        start_index = 0
        result = []
        (
            page_size_per_worker,
            refined_number_of_pages_per_worker,
        ) = _refine_page_confings(
            page_size, pages_per_worker, total_entries, number_of_workers
        )
        for worker in range(number_of_workers):
            payload = {
                "worker": worker,
                "start_index": start_index,
                "number_of_pages": refined_number_of_pages_per_worker,
                "page_size": page_size_per_worker,
            }
            start_index += page_size * pages_per_worker
            result.append(payload)
        if page_size >= total_entries:
            result[0]["number_of_pages"] = 1
            result[0]["page_size"] = total_entries
        elif page_size * number_of_workers > total_entries:
            _refine_result_list(number_of_workers, page_size, result, total_entries)
        return result

    def _get_total_entries(self) -> int:
        query = self.query
        result = next(query.Pager(self._get_service()))
        if "totalNumEntries" in result:
            return result["totalNumEntries"]
        return 0


class GoogleAdsCustomerUtil(AdWordsDataReader):
    """
    Google ads Utility to handle customer details, customer here meaning, the Google ads customer
    account, ie. the accounts that you see on top right corner on the Google ads console
    Args:
        conn (GoogleAdsConnectionManager): Connection manager to handle the creation of
        Google ads client
    """

    def __init__(self, conn: GoogleAdsConnectionManager):
        super().__init__(conn, service="CustomerService", version="v8")

    def get_customers(self) -> List[dict]:
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]

        """
        return self._get_service().list_accessible_customers()


def _find_and_append_data(fail: dict, data: List[dict]) -> dict:
    for element in fail["fieldPathElements"]:
        if element["field"] == "operations":
            fail_index = element["index"]
            fail["data"] = data[fail_index]
            return fail
    return fail


class GoogleAdsClicksConversionUtil(GoogleAdsUtil):
    """
    Google ads Utility to handle clicks conversion
    Args:
        conn (GoogleAdsConnectionManager): Connection manager to handle the creation of
        Google ads client

    """

    def __init__(self, conn: GoogleAdsConnectionManager):
        super().__init__(conn, type="ClickConversion", version="v8")

    def click_conversion(self, data):
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]

        """
        click_conversion = self._get_type()
        click_conversion.gclid = data["gclid"]
        click_conversion.conversion_value = data["conversion_value"]
        click_conversion.conversion_date_time = data["conversion_date_time"]
        click_conversion.conversion_action = data["conversion_action"]
        click_conversion.currency_code = data["currency_code"]
        return click_conversion


class GoogleAdsUploadClickConversionsRequestUtil(GoogleAdsUtil):
    """
    Google ads Utility to upload click conversions
    Args:
        conn (GoogleAdsConnectionManager): Connection manager to handle the creation of
        Google ads client

    """

    def __init__(self, conn: GoogleAdsConnectionManager):
        super().__init__(conn, type="UploadClickConversionsRequest", version="v8")

    def upload_click_conversion(self, customer_id: str, click_conversion):
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]

        """
        request = self._get_type()
        request.customer_id = customer_id
        request.conversions.append(click_conversion)
        request.partial_failure = True
        return request


class GoogleAdsConversionActionUtil(GoogleAdsUtil):
    """
     Google ads Utility to get conversion action
    Args:
        conn (GoogleAdsConnectionManager): Connection manager to handle the creation of
        Google ads client

    """

    def __init__(self, conn: GoogleAdsConnectionManager):
        super().__init__(conn, service="ConversionActionService", version="v8")

    def get_conversion_action(self, customer_id: str, conversion_action_id: str):
        """
        Gets the conversion action of the googlea ds accounts associated with the connection
        Returns: List[dict]

        """
        return self._get_service().conversion_action_path(
            customer_id, conversion_action_id
        )


class GoogleAdsOfflineConversionUtil(GoogleAdsUtil):
    """
    Google Ads Utility to handle offline conversions
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdsConnectionManager):
        super().__init__(conn, service="ConversionUploadService", version="v8")
        self.required_fields = [
            "gclid",
            "conversion_action",
            "conversion_date_time",
            "conversion_value",
        ]
        self.valid_fields = self.required_fields + [
            "currency_code",
            "externalAttributionCredit",
            "externalAttributionModel",
        ]
        self.conversion_time_format = "%Y%m%d %H%M%S %Z"

    def test_upload_conversion(self):
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]

        """
        return self._get_service()

    def upload_conversions(self, data):
        """
        Upload a list of conversions as a batch
        Args:
            data List[dict]: a List of conversion dictionaries like:
            [
                {
                    'googleClickId': 'some valid gclid',
                    'conversionName': 'some conversion name',
                    'conversionTime': '20200309 074353 UTC',
                    'conversionValue': 17.0,
                    'conversionCurrencyCode': 'AUD',
                },
                {
                    'googleClickId': 'some other valid gclid',
                    'conversionName': 'some conversion name',
                    'conversionTime': '20200309 023001 UTC',
                    'conversionValue': 17.0,
                    'conversionCurrencyCode': 'AUD',
                    'externalAttributionCredit': 0.3,
                    'externalAttributionModel': 'Linear',
                },
            ]

        Returns (List[dict], List[dict]): A tuple of lists holding succeeded and failed records,
        respectively eg.
        (
            [
                {
                    'googleClickId': 'gclid that succeeded',
                    'conversionName': 'some conversions name',
                    'conversionTime': '20200309 074353 UTC',
                    'conversionValue': 17.0,
                    'conversionCurrencyCode': 'AUD',
                    'externalAttributionCredit': None,
                    'externalAttributionModel': None
                },
            ] ,
             [
                 {
                    'fieldPath': 'operations[0].operand',
                    'fieldPathElements': [
                        {
                            'field': 'operations',
                            'index': 0
                        },
                        {
                            'field': 'operand',
                            'index': None
                        }
                    ],
                    'trigger': None,
                    'errorString': 'OfflineConversionError.UNPARSEABLE_GCLID',
                    'ApiError.Type': 'OfflineConversionError',
                    'reason': 'UNPARSEABLE_GCLID',
                    'data': {
                        'googleClickId': 'Gclid that failed',
                        'conversionName': 'some conversion name',
                        'conversionTime': '20200309 074353 UTC',
                        'conversionValue': 17.0,
                        'conversionCurrencyCode': 'AUD',
                        'externalAttributionCredit': None,
                        'externalAttributionModel': None
                    }
                },
            ]
        )

        """
        if not data:
            return [], []
        result = self._upload_mutations_batch(data)
        partial_failure = getattr(result, "partial_failure_error", None)
        uploaded = getattr(result, "results", None)
        return uploaded, partial_failure

    def _upload_mutations_batch(self, mutations):
        return self._get_service().upload_click_conversions(mutations)


def get_page_as_list_of_dict(page: dict) -> List[OrderedDict]:
    """
    Converts a list of entries from google adwords response into a list of Ordered Dictionaries
    Args:
        page (dict): the response page from google adwords api
    Returns: List[dict]
    """
    result = []
    if "entries" in page:
        entries = page["entries"]
        # These entries are a list of zeep Objects that need conversion to Dict
        result = [zeep_object_to_dict(entry) for entry in entries]
        LOG.debug("The result from the adword API: %s", result)
    else:
        LOG.info("No entries were found.")
    return result


def zeep_object_to_dict(obj: Any) -> OrderedDict:
    """
    converts the zeep client objects used in google adwords to Ordered Dict, also converts nested
    objects to ordered dict recursively
    Args:
        obj (Any): the Zeep object to be converted
    Returns: OrderedDict
    """
    obj_dict = obj.__dict__["__values__"]
    for key, val in obj_dict.items():
        if "zeep.objects" in f"{type(val)}":
            obj_dict[key] = zeep_object_to_dict(val)
    return obj_dict


class AdWordsManagedCustomerUtil(GoogleAdsUtil):
    """
    Adwords Utility to parse through and gather Account Information for all sub accounts
    Args:
        conn (GoogleAdWordsConnectionManager): Connection manager to handle the creation of
        adwords client
    """

    def __init__(self, conn: GoogleAdsConnectionManager):
        super().__init__(conn, service="ManagedCustomerService", version="v201809")
        self.selector_fields = [
            "Name",
            "CustomerId",
            "DateTimeZone",
            "CurrencyCode",
            "CanManageClients",
            "TestAccount",
            "AccountLabels",
        ]

    def _init_selector(self, offset, page_size):
        return {
            "fields": self.selector_fields,
            "paging": {"startIndex": str(offset), "numberResults": str(page_size)},
        }

    def get_all_accounts(self, page_size: int = 1000) -> List[dict]:
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: List[dict]
        """
        # Construct selector to get all accounts.
        offset = 0
        selector = self._init_selector(offset, page_size)
        more_pages = True
        all_accounts = []
        while more_pages:
            # Get serviced account graph.
            page = self._get_service().get(selector)
            all_accounts.extend(get_page_as_list_of_dict(page))
            offset += page_size
            selector["paging"]["startIndex"] = str(offset)
            more_pages = offset < int(page["totalNumEntries"])

        return all_accounts

    def get_all_accounts_as_dataframe(self, page_size: int = 1000) -> DataFrame:
        """
        Gets the customer details of the adwords accounts associated with the connection
        Returns: DataFrame
        """
        return nested_list_of_dict_to_dataframe(
            self.get_all_accounts(page_size=page_size)
        )



"""
This module helps us version track classes by capturing the
latest git hash of the file they are written in, and dumping these versions to
a json file

The application looks through a package for classes with the registration
decorator. The last git commit has for files which contain this decorator are
then used to track the current version of the class

Example:
  The application takes two arguments the name of the package you're interested
  in, and the output location for the json file

  $ python version-tracker -p transformers -o ./version.json

It also provides transparent functions which can be used as decorators to
classes, functions and methods to tag them as requiring tracking by the
version control app

Typical usage:

from version_tracking import register_class_for_version_tracking
from version_tracking import register_method_for_version_tracking

@register_class_for_version_tracking
class Foo:
    def __init__(self):
        self.foo = 'some value'


    @register_method_for_version_tracking
    def track_this_method(self):
        return self.foo

When the version tracking app is executed the latest githash of the files
containing registered classes, functions, and methods will be logged to disk
in a json file
"""

import argparse
import json
import os
import socket
import traceback
from datetime import datetime
from pathlib import Path

import git
import joblib

from hip_data_tools.common import LOG as log

CLASS_DECORATOR_STRING = '@register_class_for_version_tracking'
"string : Used to trigger identification of the class by the version tracker"

METHOD_DECORATOR_STRING = '@register_method_for_version_tracking'
"string : Used to trigger identification of the method by the version tracker"

DEFINITION_MAPPING = {
    'class': CLASS_DECORATOR_STRING,
    'def': METHOD_DECORATOR_STRING
}
"dictionary: Used to map between the type and their respective decorator"


def register_class_for_version_tracking(cls):
    """
    Transparent decorator class used for registering a class with version
    monitoring.

    Args:
        cls: class to decorate

    Returns: decorated class

    """
    return cls


def register_method_for_version_tracking(meth):
    """
    Transparent decorator method used for registering a class with version
    monitoring.

    Args:
        meth: class to decorate

    Returns: decorated class

    """
    return meth


def _find_package_location(pkg):
    """
    find the location on disk of the package of interest
    Args:
        pkg: package of interest

    Returns: string, folder location of the package

    """
    loc = _get_package_location(pkg)
    return loc.split("__init__.py")[0]


def _get_package_location(pkg):
    """
    Finds the location of the __init__.py file of the package
    Args:
        pkg: package of interest

    Returns: string, the location of the __init__.py file of the package

    """
    return pkg.__file__


def _find_all_python_modules_in_folder(fld):
    """
    Finds all python modules within a given folder and its subfolders
    Args:
        fld: folder of interest

    Returns: list of paths to the python files

    """
    return list(Path(fld).rglob("*.py"))


def _instantiate_repo(repo_path):
    """
    Instaniates a git repo job at the the given path
    Args:
        repo_path: absolute or relative path the github repo

    Returns: github repo object

    """
    return git.Repo(repo_path)


def find_tracked_modules(file_list):
    """
    Iterates through a list of python files, and returns the names of classes,
    methods and functions and their file locations, which require a version
    to be found
    Args:
        file_list: list of paths to python files

    Returns:
            definitions_with_tags: list of class names which require version
            tracking
            files_with_tag: list of files in which the version tracked files
                are found

    """
    files_with_tag = []
    definitions_with_tags = []

    for path in file_list:
        defs_with_tags_in_file = \
            find_any_relevant_decorations_in_file(path)

        definitions_with_tags = definitions_with_tags + defs_with_tags_in_file

        if defs_with_tags_in_file:
            files_with_tag.extend(
                [path for _ in defs_with_tags_in_file]
            )

    return definitions_with_tags, files_with_tag


def find_any_relevant_decorations_in_file(path):
    """
    Find any relevant decorations in the provided in tehe
    Args:
        path (string): Path to the file being scanned

    Returns (list(string)): List of decorated definitions found in the file

    """

    definitions_with_tags = []

    for declaration, decorating_string in DEFINITION_MAPPING.items():
        definitions_with_tags.extend(
            check_for_decorated_declaration_in_file(
                path,
                decorating_string,
                declaration))

    return definitions_with_tags


def check_for_decorated_declaration_in_file(path,
                                            decorating_string,
                                            declaration):
    """
    Finds any classes in a file with the relevant decorator string
    Args:
        path: absolute or relative path to the python file
        decorating_string (string) : string of decorator used to identify the
             methods, classes and functions which need tracked
        declaration (string): Declaration to look for
    Returns: list of classes within the file which require version tracking

    """

    lines_in_file = _load_file_lines_into_list(path)

    lines_with_decorator = _find_lines_with_decorator(decorating_string,
                                                      lines_in_file)

    tagged_declarations = _find_decorated_declarations(declaration,
                                                       lines_in_file,
                                                       lines_with_decorator)

    return tagged_declarations


def _find_decorated_declarations(declaration,
                                 lines_in_file,
                                 lines_with_decorator):
    tagged_declarations = []
    line_limit = len(lines_in_file) - 1
    for decorator_indicies in lines_with_decorator:
        tagged_declarations.extend(
            _check_lines_after_declaration(declaration,
                                           decorator_indicies,
                                           lines_in_file,
                                           line_limit))

    return tagged_declarations


def _check_lines_after_declaration(declaration,
                                   decorator_indicies,
                                   lines_in_file,
                                   line_limit):
    tagged_declarations = []
    for cur_line_index in range(decorator_indicies, line_limit + 1):
        cur_line = lines_in_file[cur_line_index]
        found_value = _check_line_for_declaration(cur_line,
                                                  declaration,
                                                  cur_line_index >= line_limit,
                                                  decorator_indicies)

        if found_value is not None:
            tagged_declarations.append(found_value)
            break

    return tagged_declarations


def _check_line_for_declaration(cur_line,
                                declaration,
                                end_of_file,
                                decorator_indicies):
    declaration_in_line = cur_line.startswith(declaration + " ")
    # Ignore the line if its empty string indicating whitespace,
    # or if its another decorator
    # and stop if its the end of the file
    if (not cur_line or not cur_line.startswith("@")) \
        and end_of_file and not declaration_in_line:
        raise DecoratorError(declaration, cur_line, decorator_indicies)

    out = None
    if declaration_in_line:
        out = _extract_declaration_name(declaration, cur_line)

    return out


def _find_lines_with_decorator(decorating_string, lines_in_file):
    # find index of lines with decorator
    lines_with_decorator = []
    for idx, line in enumerate(lines_in_file):
        if line.startswith(decorating_string):
            lines_with_decorator.append(idx)
    return lines_with_decorator


def _load_file_lines_into_list(path):
    with open(path, 'r') as file:
        lines_in_file = [line.strip() for line in file.readlines()]
    return lines_in_file


def _extract_declaration_name(declaration, line):
    declaration_name = line.strip().split(declaration)[1] \
        .split("(")[0] \
        .replace(":", "") \
        .strip()
    return declaration_name


def get_latest_git_hash_of_files_in_repo(repo, files_to_get):
    """
    Iterates through a files and gets their latest commit hash using the repo
    object to interface with git
    Args:
        repo: git Repo object
        files_to_get: list of absolute or relative paths of files which
            require the git hashes

    Returns: list of git hashes for the file paths supplied

    """
    git_hashes = []

    for path in files_to_get:
        commit = next(repo.iter_commits(paths=path, max_count=1))
        git_hashes.append(commit.hexsha)

    return git_hashes


def find_relevant_file_versions(package_location, repo_location):
    """
    Finds all the classes, and versions, for classes with the relevant
    decorator given the package and the repo location
    Args:
        package_location (str): path to package
        repo_location (str): path to relevant git repo

    Returns: dictionary of classes and the latest git hash of the
        file in which they live

    """

    module_locations = _find_all_python_modules_in_folder(package_location)
    classes_with_tag, files_with_tag = \
        find_tracked_modules(module_locations)
    repo = _instantiate_repo(repo_location)
    git_hashes = get_latest_git_hash_of_files_in_repo(repo, files_with_tag)

    version_dict = \
        {class_name: git_hash
         for class_name, git_hash in zip(classes_with_tag, git_hashes)}
    return version_dict


def write_versions_to_json(version_dict, output_location):
    """
    Creates a json of the versions dictionary and writes it to file
    Args:
        version_dict: version dictionary of classes and the latest git
        commit hash of the file in which they live
        output_location: location to write output json file to

    """
    with open(output_location, 'w') as file:
        json.dump(version_dict, file)


def find_and_export_relevant_versions(path,
                                      repo_location,
                                      output_location):
    """
    Looks through a package and finds classes which have been decorated
    as requiring version tagging, and then writes these classes and their
    versions to a json file
    Args:
        path: local path to package
        repo_location: location of github repo used to track versions
        output_location: location of version control json report

    """

    versions_dict = find_relevant_file_versions(path, repo_location)

    write_versions_to_json(versions_dict, output_location)


class DecoratorError(Exception):
    """
    Error raised when there is a problem with the decoration of a class or
    method

    Args:
        declaration (str): Declartion the application is looking for
        line (str): Line where the declaration is made
        line_number (int): Line number in found of declaration

    """

    def __init__(self, declaration, line, line_number):
        # Call the base class constructor with the parameters it needs
        message_body = "Decorator Found, but no declaration discovered. " \
                       "Declaration {}, at line {}, line number {}" \
            .format(declaration, line, str(line_number))

        super().__init__(message_body)


class VersionTracker:
    """
    A class that allows us to track versions through hashes and strings by
    generating a dictionary. Versions can be loaded from file or can be
    generated on the fly from object hashes

    """

    def __init__(self):

        self._version_dict = {}

    def add_dictionary_to_versions(self, in_dict):
        """
        Add a Python dictionary to the versioning dictionary
        Args:
            in_dict (dict): A dictionary to append to the version dictionary

        Returns: None

        """

        self._version_dict.update(in_dict)

    def add_versions_from_json_file(self, version_file_location):
        """
        Load a json file from local storage and append these versions to the
        version tracking dictionary
        Args:
            version_file_location (str): Path to the json file

        Returns: None

        """
        try:
            with open(version_file_location, "r") as content:
                file_content = content.read()
                version_dict_from_file = json.loads(file_content)

                self.add_dictionary_to_versions(version_dict_from_file)

        except FileNotFoundError as fnf:
            log.error("No versioning file found at : %s",
                      version_file_location)
            raise fnf

        except json.decoder.JSONDecodeError as decode_error:

            log.error("JSON file failed to decode, check version dict is "
                      "correctly formatted")

            raise decode_error

        except Exception as exception:
            log.error("unknown error")
            log.error(traceback.format_exc())
            raise exception

    def _add_hostname(self):

        self._version_dict.update({'hostname': socket.gethostname()})

    def add_object_version(self, dict_key, obj):
        """Adds a hash of an object to the version tracking
        Args:
            dict_key (str): Name to be using in the version tracking
                dictioanry
            obj (Object): Any hashable object

        Returns: None
        """

        self.add_dictionary_to_versions({dict_key: joblib.hash(obj)})

    def add_string_to_version_tracking(self, dict_key, in_string):
        """
        Adds a string value to the version tracking.
        Args:
            dict_key (str): Name to use as the key in the version tracking
                dictionary
            in_string (str): Value to be stored in the version tracking
                dictionary

        Returns: None

        """

        self.add_dictionary_to_versions({dict_key: in_string})

    def _set_agglomerated_version_hash(self):
        """
        This hash takes all of the versioning hashes and combines them to
        produce a single hash of everything excluding the timestamp and
        hostname. This should, therefore, provide a single hash which
        encapsulates the software versions.

        Returns: None

        """
        relevant_versions = {x: self._version_dict[x]
                             for x in self._version_dict if x not in
                             ['hostname', 'versioning_timestamp']}

        version_hash = joblib.hash(json.dumps(relevant_versions,
                                              sort_keys=True)
                                   )

        self.add_dictionary_to_versions({'aggregated_version': version_hash})

    def _add_versioning_timestamp(self):
        self.add_dictionary_to_versions({'versioning_timestamp':
                                             datetime.utcnow().isoformat()})

    def get_version_dict(self):
        """
        Brings together all of the required versioning information and
        returns the relevant dictionary
        Returns (dict): Dictionary which tracks all of the tracked versions
        """

        self._set_agglomerated_version_hash()
        self._add_hostname()
        self._add_versioning_timestamp()

        return self._version_dict


def main():
    """
    Main function which takes command line arguments
        and executes the version finder
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("-p",
                        "--package",
                        dest='pkg_nm',
                        required=True,
                        metavar='str',
                        help="package to be inspected")

    parser.add_argument("-r",
                        "--repo",
                        dest='repo',
                        required=False,
                        default=".",
                        metavar='str',
                        help="location of git repo")

    parser.add_argument("-o",
                        "--output-location",
                        dest='out_loc',
                        required=False,
                        default="version_tracking.json",
                        metavar='str',
                        help="location to put versioning file")

    parser.add_argument("-d",
                        "--pacakge-directory",
                        dest='pkg_dir',
                        required=False,
                        default=os.getcwd(),
                        metavar='str',
                        help="location to put versioning file")

    args = parser.parse_args()

    find_and_export_relevant_versions(args.pkg_nm, args.repo, args.out_loc)


if __name__ == "__main__":
    main()



"""
Module to deal with data transfer from S3 to Cassandra
"""
from typing import List, Optional

from hip_data_tools.etl.common import ETL
from hip_data_tools.etl.s3 import S3SourceSettings, S3SinkSettings, S3FileLocationExtractor, \
    S3FileCopyLoader, S3Key, S3FileNameTransformer, AddTargetS3KeyTransformer


class S3ToS3FileCopy(ETL):
    """
    Class to transfer objects from s3 to s3
    Args:
        source (S3FileLocationExtractor):
        sink (S3FileCopyLoader):
    Eg:
    >>> etl = S3ToS3FileCopy(
    ...     source = S3SourceSettings(
    ...         bucket="MY_SOURCE_BUCKET_NAME",
    ...         key_prefix="foo/bar/",
    ...         suffix="parquet",
    ...         connection_settings=aws_setting,
    ...     ),
    ...     sink = S3SinkSettings(
    ...         bucket="MY_TARGET_BUCKET_NAME",
    ...         connection_settings=aws_setting,
    ...     ),
    ...     transformers = [AddTargetS3KeyTransformer(target_key_prefix="bar/baz/")],
    ... )
    ...
    >>>
    >>> etl.execute_next()
    ...

    Args:
        source (S3SourceSettings): Settings for source s3 files
        sink (S3SinkSettings): Settings for target s3 directory
        transformers (List[S3FileNameTransformer]): Transformers to change file names
    """

    def __init__(self, source: S3SourceSettings, sink: S3SinkSettings,
                 transformers: Optional[List[S3FileNameTransformer]] = None):
        default_transformer = AddTargetS3KeyTransformer(target_key_prefix=None)
        if not transformers:
            transformers = [default_transformer]
        super().__init__(
            extractor=S3FileLocationExtractor(source),
            loader=S3FileCopyLoader(sink),
            transformers=transformers,
        )

    def list_source_files(self) -> List[S3Key]:
        """
        List the source files as per the source settings

        Returns: List[S3Key]

        """
        return self.extractor.list_source_files()



"""
Module to deal with data transfer from Adwords to S3
"""
from typing import List, Optional, Dict

import numpy as np
import pandas as pd
from attr import dataclass
from googleads.adwords import ServiceQueryBuilder, ReportQuery
from pandas import DataFrame

from hip_data_tools.aws.common import AwsConnectionSettings, AwsConnectionManager
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.common import dataframe_columns_to_snake_case
from hip_data_tools.google.adwords import GoogleAdWordsConnectionSettings, AdWordsDataReader, \
    GoogleAdWordsConnectionManager, AdWordsParallelDataReadEstimator, AdWordsReportReader


@dataclass
class AdWordsToS3Settings:
    """S3 to Cassandra ETL settings"""
    source_query_fragment: ServiceQueryBuilder
    source_service: str
    source_service_version: str
    source_connection_settings: GoogleAdWordsConnectionSettings
    target_bucket: str
    target_key_prefix: str
    target_file_prefix: Optional[str]
    target_connection_settings: AwsConnectionSettings


class AdWordsToS3:
    """
    ETL Class to handle the transfer of data from adwords based on AWQL to S3 as parquet files
    Args:
        settings (AdWordsToS3Settings): the etl settings to be used
    """

    def __init__(self, settings: AdWordsToS3Settings):
        self.__settings = settings
        self._adwords_util = None
        self._s3_util = None
        self._source_keys = None
        self.start_index = None
        self.page_size = None
        self.query = None
        self.iteration_limit = 0
        self.current_iteration = 0

    def _get_s3_util(self) -> S3Util:
        if self._s3_util is None:
            self._s3_util = S3Util(
                bucket=self.__settings.target_bucket,
                conn=AwsConnectionManager(self.__settings.target_connection_settings),
            )
        return self._s3_util

    def _get_adwords_util(self) -> AdWordsDataReader:
        if self._adwords_util is None:
            self._adwords_util = AdWordsDataReader(
                conn=GoogleAdWordsConnectionManager(self.__settings.source_connection_settings),
                service=self.__settings.source_service,
                version=self.__settings.source_service_version
            )
        return self._adwords_util

    def build_query(self, start_index: int, page_size: int, num_iterations: int) -> None:
        """
        Builds the query based on the query fragment in settings, to be able to load data in
        parallel from the given start index for a number of iterations
        Args:
            start_index (int): the start index to offset the beginning of query paging
            page_size (int): number of elements in each page/api call
            num_iterations (int): total number of pages required for transfer of entire data
        Returns: None
        """
        query_fragment = self.__settings.source_query_fragment
        self.start_index = start_index
        self.page_size = page_size
        self.query = query_fragment.Limit(start_index=self.start_index, page_size=page_size).Build()
        self.iteration_limit = num_iterations
        self.current_iteration = 0
        self._get_adwords_util().set_query(self.query)

    def transfer_next_iteration(self) -> bool:
        """
        Transfers the next page of data
        Returns: bool true if the data transfer succeeded, False if reached end of iterations
        """
        if self.current_iteration < self.iteration_limit:
            return self.__upload_next_page_data()
        else:
            return False

    def __upload_next_page_data(self):
        try:
            data = self._get_next_page()
        except StopIteration:
            return False
        if data.empty:
            return False
        s3u = self._get_s3_util()
        s3u.upload_dataframe_as_parquet(
            dataframe=data,
            key=self.__settings.target_key_prefix,
            file_name=self.__get_file_name())
        self.current_iteration += 1
        return True

    def __get_file_name(self):
        file_prefix_str = ""
        if self.__settings.target_file_prefix is not None:
            file_prefix_str = self.__settings.target_file_prefix
        return f"{file_prefix_str}index_{self._get_current_start_index()}__" \
            f"{self._get_current_end_index()}"

    def _get_next_page(self) -> DataFrame:
        if not self.query:
            raise ValueError(
                "query is not set properly. please use the build_query() method to set it up.")
        data = self._get_adwords_util().download_next_page_as_dataframe()
        return data

    def transfer_all(self) -> None:
        """
        Iteratively transfer all pages of data
        Returns: None
        """
        while True:
            if not self.transfer_next_iteration():
                break

    def _get_current_start_index(self) -> int:
        return int(self.start_index + (self.page_size * self.current_iteration))

    def _get_current_end_index(self) -> int:
        return int(self._get_current_start_index() + (self.page_size - 1))

    def get_parallel_payloads(self, page_size: int, number_of_workers: int) -> List[dict]:
        """
        gives a list of dicts that contain start index, page size, and number of iterations
        Args:
            page_size (int): number of elements in each page / api call
            number_of_workers (int): total number of parallel workers for which the payload needs
            to be distributed
        Returns: List[dict] eg:
        [
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 0, 'worker': 0},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 393000, 'worker': 1},
            {'number_of_pages': 393, 'page_size': 1000, 'start_index': 786000, 'worker': 2},
        ]
        """

        estimator = AdWordsParallelDataReadEstimator(
            conn=GoogleAdWordsConnectionManager(self.__settings.source_connection_settings),
            service=self.__settings.source_service,
            version=self.__settings.source_service_version,
            query=self.__settings.source_query_fragment.Limit(0, 1).Build())
        return estimator.get_parallel_payloads(page_size, number_of_workers)


@dataclass
class AdWordsReportToS3Settings:
    """S3 to Cassandra ETL settings"""
    source_query: ReportQuery
    source_include_zero_impressions: bool
    source_connection_settings: GoogleAdWordsConnectionSettings
    target_bucket: str
    target_key_prefix: str
    target_file_prefix: Optional[str]
    target_connection_settings: AwsConnectionSettings
    transformation_field_type_mask: Optional[Dict[str, np.dtype]]


class AdWordsReportsToS3:
    """
    ETL Class to handle the transfer of data from adwords reports based on AWQL to S3 as parquet
    Args:
        settings (AdWordsToS3Settings): the etl settings to be used
    """

    def __init__(self, settings: AdWordsReportToS3Settings):
        self.__settings = settings
        self._adwords_util = None
        self._s3_util = None

    def _get_s3_util(self) -> S3Util:
        if self._s3_util is None:
            self._s3_util = S3Util(
                bucket=self.__settings.target_bucket,
                conn=AwsConnectionManager(self.__settings.target_connection_settings),
            )
        return self._s3_util

    def _get_adwords_util(self) -> AdWordsReportReader:
        if self._adwords_util is None:
            self._adwords_util = AdWordsReportReader(
                conn=GoogleAdWordsConnectionManager(self.__settings.source_connection_settings))
        return self._adwords_util

    def _mask_field_types(self, df: DataFrame):
        for field_name, field_type in self.__settings.transformation_field_type_mask.items():
            if field_type is int:
                df[field_name] = pd.to_numeric(df[field_name], errors='coerce')
                df[field_name].fillna(0, inplace=True)

            df[field_name] = df[field_name].astype(field_type)

    def transfer(self, **kwargs):
        """
        Transfer the entire report to s3 in parquet format
        Returns: None
        """
        data = self._get_report_data(**kwargs)
        s3u = self._get_s3_util()
        file_name = "report_data"
        if self.__settings.target_file_prefix:
            file_name = f"{self.__settings.target_file_prefix}{file_name}"
        dataframe_columns_to_snake_case(data)
        if self.__settings.transformation_field_type_mask:
            self._mask_field_types(data)
        s3u.upload_dataframe_as_parquet(
            dataframe=data,
            key=self.__settings.target_key_prefix,
            file_name=file_name)

    def _get_report_data(self, **kwargs):
        au = self._get_adwords_util()
        data = au.awql_to_dataframe(
            self.__settings.source_query,
            self.__settings.source_include_zero_impressions,
            **kwargs)
        return data



"""
Module to deal with data transfer from Google sheets to Athena
"""

from attr import dataclass

from hip_data_tools.aws.athena import AthenaUtil, get_table_settings_for_dataframe
from hip_data_tools.aws.common import AwsConnectionManager
from hip_data_tools.etl.google_sheet_to_s3 import GoogleSheetToS3, GoogleSheetsToS3Settings


@dataclass
class GoogleSheetsToAthenaSettings(GoogleSheetsToS3Settings):
    """
    Google sheets to Athena ETL settings
    Args:
        source_workbook_url: the url of the workbook
            (eg: https://docs.google.com/spreadsheets/d/1W1vIBLfsQM/edit?usp=sharing)
        source_sheet: name of the google sheet (eg: sheet1)
        source_row_range: range of rows (eg: '2:5')
        source_field_names_row_number: row number of the field names (eg: 4). Assumes the data
            starts at first column and  there is no gaps. There should not be 2 fields with the same
            name.
        source_field_types_row_number: row number of the field types (eg: 5)
        source_data_start_row_number: starting row number of the actual data
        source_connection_settings: GoogleApiConnectionSettings with google api keys dictionary
            object
        manual_partition_key_value: a dictionary with partition column name and value. Only one
            partition key can be used and this value need to be string
            (eg: {"column": "start_date", "value": "2020-03-08"})
        target_database: name of the athena database (eg: dev)
        target_table_name: name of the athena table (eg: 'sheet_table')
        target_s3_bucket: s3 bucket to store the files (eg: au-test-bucket)
        target_s3_dir: s3 directory to store the files (eg: sheets/new)
        target_connection_settings: aws connection settings
        target_table_ddl_progress: if this is true, the target table will be dropped and recreated
    """
    target_database: str
    target_table_name: str
    target_table_ddl_progress: bool


class GoogleSheetToAthena(GoogleSheetToS3):
    """
    Class to transfer data from google sheet to athena
    Args:
        settings (GoogleSheetsToAthenaSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: GoogleSheetsToAthenaSettings):
        self.__settings = settings
        self.base_s3_dir = self.__settings.target_s3_dir
        self.__settings.target_s3_dir = self._calculate_s3_key()
        super().__init__(self.__settings)
        self.keys_to_transfer = None

    def _get_athena_util(self):
        return AthenaUtil(database=self.__settings.target_database,
                          conn=AwsConnectionManager(
                              settings=self.__settings.target_connection_settings),
                          output_bucket=self.__settings.target_s3_bucket)

    def load_sheet_to_athena(self):
        """
        Load google sheet into Athena
        :return: None
        """
        self.write_sheet_data_to_s3()

        athena_util = self._get_athena_util()
        if self.__settings.target_table_ddl_progress:
            athena_util.drop_table(self.__settings.target_table_name)

        athena_util.create_table(table_settings=get_table_settings_for_dataframe(
            dataframe=self._get_sheet_dataframe(),
            partitions=self.__settings.manual_partition_key_value,
            table=self.__settings.target_table_name,
            s3_bucket=self.__settings.target_s3_bucket,
            s3_dir=self.base_s3_dir))

        if self.__settings.target_table_ddl_progress:
            athena_util.repair_table_partitions(table=self.__settings.target_table_name)
        else:
            athena_util.add_partitions(
                table=self.__settings.target_table_name,
                partition_keys=[self.__settings.manual_partition_key_value["column"]],
                partition_values=[self.__settings.manual_partition_key_value["value"]]
            )

    def _calculate_s3_key(self):
        s3_key_with_partition = self.__settings.target_s3_dir
        if self.__settings.manual_partition_key_value is not None:
            column_name = self.__settings.manual_partition_key_value["column"]
            column_value = self.__settings.manual_partition_key_value["value"]
            partition_path = f"/{column_name}={column_value}"
            s3_key_with_partition += partition_path
        return s3_key_with_partition



"""
Module to deal with data transfer from S3 to Cassandra
"""
from typing import Optional, List

from attr import dataclass
from cassandra.datastax.graph import Result

from hip_data_tools.apache.cassandra import CassandraUtil, CassandraConnectionManager, \
    CassandraConnectionSettings
from hip_data_tools.aws.common import AwsConnectionManager
from hip_data_tools.aws.s3 import S3Util
from hip_data_tools.common import LOG
from hip_data_tools.etl.s3_to_dataframe import S3ToDataFrame, S3ToDataFrameSettings


@dataclass
class S3ToCassandraSettings(S3ToDataFrameSettings):
    """S3 to Cassandra ETL settings"""
    destination_keyspace: str
    destination_table: str
    destination_table_primary_keys: List[str]
    destination_table_partition_key: Optional[List[str]]
    destination_connection_settings: CassandraConnectionSettings
    destination_table_options_statement: str = ""
    destination_batch_size: int = 1


class S3ToCassandra(S3ToDataFrame):
    """
    Class to transfer parquet data from s3 to Cassandra
    Args:
        settings (S3ToCassandraSettings): the settings around the etl to be executed
    """

    def __init__(self, settings: S3ToCassandraSettings):
        self.__settings = settings
        super().__init__(self.__settings)
        self.keys_to_transfer = None

    def _get_cassandra_util(self):
        return CassandraUtil(
            keyspace=self.__settings.destination_keyspace,
            conn=CassandraConnectionManager(
                settings=self.__settings.destination_connection_settings),
        )

    def _get_s3_util(self):
        return S3Util(
            bucket=self.__settings.source_bucket,
            conn=AwsConnectionManager(self.__settings.source_connection_settings),
        )

    def create_table(self):
        """
        Creates the destination cassandra table if not exists
        Returns: None
        """
        files = self.list_source_files()
        data_frame = self._get_s3_util().download_parquet_as_dataframe(
            key=files[0])
        # use specified partition and clustering keys
        self._get_cassandra_util().create_table_from_dataframe(
            data_frame=data_frame,
            table_name=self.__settings.destination_table,
            primary_key_column_list=self.__settings.destination_table_primary_keys,
            partition_key_column_list=self.__settings.destination_table_partition_key,
            table_options_statement=self.__settings.destination_table_options_statement,
        )

    def create_and_upsert_all(self) -> List[List[Result]]:
        """
        First creates the table and then upsert all s3 files to the table
        Returns: None
        """
        self.create_table()
        return self.upsert_all_files()

    def upsert_all_files(self) -> List[List[Result]]:
        """
        Upsert all files from s3 sequentially into cassandra
        Returns: None
        """
        return [self._upsert_data_frame(df) for df in self._get_iterator()]

    def _upsert_data_frame(self, data_frame):
        if self.__settings.destination_batch_size > 1:
            LOG.info("Going to upsert batches of size %s", self.__settings.destination_batch_size)
            result = self._get_cassandra_util().upsert_dataframe_in_batches(
                dataframe=data_frame,
                table=self.__settings.destination_table,
                batch_size=self.__settings.destination_batch_size)
        else:
            LOG.info("Going to upsert one row at a time")
            result = self._get_cassandra_util().upsert_dataframe(
                dataframe=data_frame,
                table=self.__settings.destination_table)
        return result

    def upsert_file(self, key: str) -> List[Result]:
        """
        Read a parquet file from s3 and upsert the records to Cassandra
        Args:
            key: s3 key for the parquet file
        Returns: None
        """
        data_frame = self.get_data_frame(key)
        return self._upsert_data_frame(data_frame)
